import sys
import exceptions
import time
import os
import signal
import termios

class UML:
    def __init__(self):
        self.umid = 'umlrun-%d' % os.getpid() # XXX, should guarantee uniqueness
        self.verbose = None
        self.umlpid = None
        self.control = None
        self.interactive = None
        self.version = None

    def start(self, umlargs, verbose=0, timeout=30, interactive='auto'):
        uml = ['linux',
               'umlrun=ssl0,ssl1',
               'ssl0=pts',
               'con=pts',
               'umid=' + self.umid]
        uml.extend(umlargs)
        self.verbose = verbose
        self.saved_ttyattr = None
        if sys.stdout.isatty():
            self.saved_ttyattr = termios.tcgetattr(sys.stdout.fileno())
            
        if interactive == 'auto':
            if sys.stdin.isatty() and sys.stdout.isatty():
                self.interactive = 1
            else:
                self.interactive = 0
        else:
            self.interactive = interactive

        if self.interactive:
            uml.append('ssl1=tty:/dev/tty')
        else:
            uml.append('ssl1=fd:0,fd:1')

        if not self.verbose:
            uml.append('quiet')

        self.umlpid = os.fork()
        if self.umlpid == 0:
            os.execvp(uml[0], uml)
            sys.exit(1)

        self._log("Waiting for UML to start up...")
        timeleft = timeout
        while timeleft > 0:
            try:
                status = os.waitpid(self.umlpid, os.WNOHANG)
                if status[0] == self.umlpid:
                    raise self.Error('UML failed to start')
            except OSError:
                pass
            try:
                self._mconsole('version')
                break
            except MConsoleConnectionFailed:
                pass
            time.sleep(1)
            timeleft -= 1
        if timeleft <= 0:
            raise TimeoutError('UML did not start up within %d seconds' % timeout)

        self._log("UML startup OK (pid %s)" % self.umlpid)

        control_pts = None
        timeleft = timeout
        self._log("Waiting for umlrun-server to start up...")
        while timeleft > 0:
            control_pts = self._ptspath('ssl0')
            if control_pts:
                break

            time.sleep(1)
            timeleft -= 1
        if timeleft <= 0:
            raise TimeoutError('umlrun-server did not start up within %d seconds' % timeout)

        self._log("umlrun-server startup OK")

        self.control = open(control_pts, 'r+')

        self.version = self.control.readline().rstrip()
        self._log('umlrun-server version is ' + self.version)

    def shellcommand(self, args):
        if self.interactive:
            self._command('exec_interactive', args)
        else:
            self._command('exec', args)

    def setenv(self, name, value=None):
        if value == None:
            value = os.getenv(name)
            if value == None:
                return
        self._command('setenv', name, value)

    def mounthost(self):
        if self.version < 2:
            raise InterfaceUnsupported('mounthost')

        self._command('mount_host', os.getcwd())

    def shutdown(self, timeout=30):
        if self.control:
            self._command('shutdown')
            self.control.close()

            self._log("Waiting for UML to shut down...")
            timeleft = timeout
            while timeleft > 0:
                if os.waitpid(self.umlpid, os.WNOHANG)[0] != 0:
                    break
                time.sleep(1)
                timeleft -= 1
            if timeleft <= 0:
                self._log("Halting UML via mconsole")
                self._mconsole('halt')
                timeleft = timeout
                while timeleft > 0:
                    if os.waitpid(self.umlpid, os.WNOHANG)[0] != 0:
                        break
                if timeleft <= 0:
                    self._log("Killing UML forcibly")
                    os.kill(self.umlpid, signal.SIGTERM)

            if self.saved_ttyattr:
                termios.tcsetattr(sys.stdout.fileno(), termios.TCSANOW, self.saved_ttyattr)


    def _command(self, cmd, *args):
        self.control.write('%s %s\n' % (cmd.upper(), ' '.join(args)))
        line = self.control.readline()[:-1]
        status, result = line.split(' ', 1)
        self._log('_command(%s, %s) -> %s %s' % (cmd, ', '.join(args), status, result))
        if int(status) != 200:
            raise CommandError(cmd, status, result)

    def _mconsole(self, cmd):
        f = os.popen('uml_mconsole %s %s 2>/dev/null' % (self.umid, cmd))
        output = f.read()
        status = f.close()
        if status != None:
            #self._log("mconsole status:", status)
            raise MConsoleConnectionFailed()
        if len(output) == 0:
            # XXX this happens sometimes, why?
            raise MConsoleConnectionFailed()

        #print "mconsole(%s): %s" % (cmd, output)
        return output

    def _ptspath(self, dev):
        output = self._mconsole('config ' + dev)
        status, result = output.rstrip().split(' ', 1)
        if status == 'OK':
            if result == 'pts':
                return None
            elif result[:4] == 'pts:':
                return result[4:]
            else:
                raise Error('Unexpected result from mconsole config query: ' + result)
        else:
            raise MConsoleError(status, result)

    def _log(self, msg):
        if self.verbose:
            sys.stderr.write(msg + '\r\n')

class Error(exceptions.RuntimeError):
    pass

class MConsoleConnectionFailed(Error):
    pass

class MConsoleError(Error):
    def __init__(self, status, result):
        Error.__init__(self, 'mconsole: ' + status + ' ' + result)

class CommandError(Error):
    def __init__(self, command, status, result):
        self.command = command
        self.status = status
        self.result = result
        Error.__init__(self, command + ': ' + status + ' ' + result)

class InterfaceUnsupported(Error):
    def __init__(self, command):
        self.command = command
        Error.__init__(self, 'Interface not supported by umlrun-server')

if __name__ == '__main__':
    # simple self-test
    uml = UML()
    try:
        uml.start(umlargs=['eth0=tuntap,,,192.168.10.1',
                           'ubd0=/tmp/cow,/home/mdz/.umldeb/images/unstable/image',
                           'con=xterm'],
                  verbose=0,interactive=0)
        uml.setenv('FOO', 'bar')
        #print "Executing a shell command:"
        uml.shellcommand('echo "Hello, world! (FOO=$FOO)"')
        uml.shellcommand('/bin/bash -i')
    finally:
        uml.shutdown()
        os.remove('/tmp/cow')
